// Copyright (c) 2019 Uber Technologies, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

package options

import (
	"fmt"
	"net/http"
	"time"

	clusterclient "github.com/m3db/m3/src/cluster/client"
	"github.com/m3db/m3/src/cmd/services/m3coordinator/ingest"
	dbconfig "github.com/m3db/m3/src/cmd/services/m3dbnode/config"
	"github.com/m3db/m3/src/cmd/services/m3query/config"
	"github.com/m3db/m3/src/query/api/v1/handler/prometheus"
	"github.com/m3db/m3/src/query/api/v1/handler/prometheus/handleroptions"
	"github.com/m3db/m3/src/query/cost"
	"github.com/m3db/m3/src/query/executor"
	"github.com/m3db/m3/src/query/models"
	"github.com/m3db/m3/src/query/storage"
	"github.com/m3db/m3/src/query/storage/m3"
	"github.com/m3db/m3/src/x/clock"
	"github.com/m3db/m3/src/x/instrument"
)

const defaultTimeout = 30 * time.Second

// CustomHandler allows for custom third party http handlers.
type CustomHandler interface {
	// Route is the custom handler route.
	Route() string
	// Methods is the list of http methods this handler services.
	Methods() []string
	// Handler is the custom handler itself.
	Handler(handlerOptions HandlerOptions) (http.Handler, error)
}

// HandlerOptions represents handler options.
type HandlerOptions interface {
	// CreatedAt returns the time the options were created.
	CreatedAt() time.Time

	// Storage returns the set storage.
	Storage() storage.Storage
	// SetStorage sets the set storage.
	SetStorage(s storage.Storage) HandlerOptions

	// DownsamplerAndWriter returns the set downsampler and writer.
	DownsamplerAndWriter() ingest.DownsamplerAndWriter
	// SetDownsamplerAndWriter sets the set downsampler and writer.
	SetDownsamplerAndWriter(d ingest.DownsamplerAndWriter) HandlerOptions

	// Engine returns the engine.
	Engine() executor.Engine
	// SetEngine sets the engine.
	SetEngine(e executor.Engine) HandlerOptions

	// Clusters returns the clusters.
	Clusters() m3.Clusters
	// SetClusters sets the clusters.
	SetClusters(c m3.Clusters) HandlerOptions

	// ClusterClient returns the cluster client.
	ClusterClient() clusterclient.Client
	// SetClusterClient sets the cluster client.
	SetClusterClient(c clusterclient.Client) HandlerOptions

	// Config returns the config.
	Config() config.Configuration
	// SetConfig sets the config.
	SetConfig(c config.Configuration) HandlerOptions

	// EmbeddedDbCfg returns the embedded db config.
	EmbeddedDbCfg() *dbconfig.DBConfiguration
	// SetEmbeddedDbCfg sets the embedded db config.
	SetEmbeddedDbCfg(c *dbconfig.DBConfiguration) HandlerOptions

	// TagOptions returns the tag options.
	TagOptions() models.TagOptions
	// SetTagOptions sets the tag options.
	SetTagOptions(opts models.TagOptions) HandlerOptions

	// TimeoutOpts returns the timeout options.
	TimeoutOpts() *prometheus.TimeoutOpts
	// SetTimeoutOpts sets the timeout options.
	SetTimeoutOpts(t *prometheus.TimeoutOpts) HandlerOptions

	// Enforcer returns the enforcer.
	Enforcer() cost.ChainedEnforcer
	// SetEnforcer sets the enforcer.
	SetEnforcer(e cost.ChainedEnforcer) HandlerOptions

	// FetchOptionsBuilder returns the fetch options builder.
	FetchOptionsBuilder() handleroptions.FetchOptionsBuilder
	// SetFetchOptionsBuilder sets the fetch options builder.
	SetFetchOptionsBuilder(b handleroptions.FetchOptionsBuilder) HandlerOptions

	// QueryContextOptions returns the query context options.
	QueryContextOptions() models.QueryContextOptions
	// SetQueryContextOptions sets the query context options.
	SetQueryContextOptions(o models.QueryContextOptions) HandlerOptions

	// CPUProfileDuration returns the cpu profile duration.
	CPUProfileDuration() time.Duration
	// SetCPUProfileDuration sets the cpu profile duration.
	SetCPUProfileDuration(c time.Duration) HandlerOptions

	// PlacementServiceNames returns the placement service names.
	PlacementServiceNames() []string
	// SetPlacementServiceNames sets the placement service names.
	SetPlacementServiceNames(n []string) HandlerOptions

	// ServiceOptionDefaults returns the service option defaults.
	ServiceOptionDefaults() []handleroptions.ServiceOptionsDefault
	// SetServiceOptionDefaults sets the service option defaults.
	SetServiceOptionDefaults(s []handleroptions.ServiceOptionsDefault) HandlerOptions

	// NowFn returns the now function.
	NowFn() clock.NowFn
	// SetNowFn sets the now function.
	SetNowFn(f clock.NowFn) HandlerOptions

	// InstrumentOpts returns the instrumentation optoins.
	InstrumentOpts() instrument.Options
	// SetInstrumentOpts sets instrumentation options.
	SetInstrumentOpts(opts instrument.Options) HandlerOptions
}

// HandlerOptions represents handler options.
type handlerOptions struct {
	storage               storage.Storage
	downsamplerAndWriter  ingest.DownsamplerAndWriter
	engine                executor.Engine
	clusters              m3.Clusters
	clusterClient         clusterclient.Client
	config                config.Configuration
	embeddedDbCfg         *dbconfig.DBConfiguration
	createdAt             time.Time
	tagOptions            models.TagOptions
	timeoutOpts           *prometheus.TimeoutOpts
	enforcer              cost.ChainedEnforcer
	fetchOptionsBuilder   handleroptions.FetchOptionsBuilder
	queryContextOptions   models.QueryContextOptions
	instrumentOpts        instrument.Options
	cpuProfileDuration    time.Duration
	placementServiceNames []string
	serviceOptionDefaults []handleroptions.ServiceOptionsDefault
	nowFn                 clock.NowFn
}

// EmptyHandlerOptions returns  default handler options.
func EmptyHandlerOptions() HandlerOptions {
	return &handlerOptions{
		instrumentOpts: instrument.NewOptions(),
		nowFn:          time.Now,
	}
}

// NewHandlerOptions builds a handler options.
func NewHandlerOptions(
	downsamplerAndWriter ingest.DownsamplerAndWriter,
	tagOptions models.TagOptions,
	engine executor.Engine,
	m3dbClusters m3.Clusters,
	clusterClient clusterclient.Client,
	cfg config.Configuration,
	embeddedDbCfg *dbconfig.DBConfiguration,
	enforcer cost.ChainedEnforcer,
	fetchOptionsBuilder handleroptions.FetchOptionsBuilder,
	queryContextOptions models.QueryContextOptions,
	instrumentOpts instrument.Options,
	cpuProfileDuration time.Duration,
	placementServiceNames []string,
	serviceOptionDefaults []handleroptions.ServiceOptionsDefault,
) (HandlerOptions, error) {
	var timeoutOpts = &prometheus.TimeoutOpts{}
	if embeddedDbCfg == nil || embeddedDbCfg.Client.FetchTimeout == nil {
		timeoutOpts.FetchTimeout = defaultTimeout
	} else {
		timeout := *embeddedDbCfg.Client.FetchTimeout
		if timeout <= 0 {
			return nil,
				fmt.Errorf("m3db client fetch timeout should be > 0, is %d", timeout)
		}

		timeoutOpts.FetchTimeout = timeout
	}

	return &handlerOptions{
		storage:               downsamplerAndWriter.Storage(),
		downsamplerAndWriter:  downsamplerAndWriter,
		engine:                engine,
		clusters:              m3dbClusters,
		clusterClient:         clusterClient,
		config:                cfg,
		embeddedDbCfg:         embeddedDbCfg,
		createdAt:             time.Now(),
		tagOptions:            tagOptions,
		timeoutOpts:           timeoutOpts,
		enforcer:              enforcer,
		fetchOptionsBuilder:   fetchOptionsBuilder,
		queryContextOptions:   queryContextOptions,
		instrumentOpts:        instrumentOpts,
		cpuProfileDuration:    cpuProfileDuration,
		placementServiceNames: placementServiceNames,
		serviceOptionDefaults: serviceOptionDefaults,
		nowFn:                 time.Now,
	}, nil
}

func (o *handlerOptions) CreatedAt() time.Time {
	return o.createdAt
}

func (o *handlerOptions) Storage() storage.Storage {
	return o.storage
}

func (o *handlerOptions) SetStorage(s storage.Storage) HandlerOptions {
	opts := *o
	opts.storage = s
	return &opts
}

func (o *handlerOptions) DownsamplerAndWriter() ingest.DownsamplerAndWriter {
	return o.downsamplerAndWriter
}

func (o *handlerOptions) SetDownsamplerAndWriter(
	d ingest.DownsamplerAndWriter) HandlerOptions {
	opts := *o
	opts.downsamplerAndWriter = d
	return &opts
}

func (o *handlerOptions) Engine() executor.Engine {
	return o.engine
}

func (o *handlerOptions) SetEngine(e executor.Engine) HandlerOptions {
	opts := *o
	opts.engine = e
	return &opts
}

func (o *handlerOptions) Clusters() m3.Clusters {
	return o.clusters
}

func (o *handlerOptions) SetClusters(c m3.Clusters) HandlerOptions {
	opts := *o
	opts.clusters = c
	return &opts
}

func (o *handlerOptions) ClusterClient() clusterclient.Client {
	return o.clusterClient
}

func (o *handlerOptions) SetClusterClient(
	c clusterclient.Client) HandlerOptions {
	opts := *o
	opts.clusterClient = c
	return &opts
}

func (o *handlerOptions) Config() config.Configuration {
	return o.config
}

func (o *handlerOptions) SetConfig(c config.Configuration) HandlerOptions {
	opts := *o
	opts.config = c
	return &opts
}

func (o *handlerOptions) EmbeddedDbCfg() *dbconfig.DBConfiguration {
	return o.embeddedDbCfg
}

func (o *handlerOptions) SetEmbeddedDbCfg(
	c *dbconfig.DBConfiguration) HandlerOptions {
	opts := *o
	opts.embeddedDbCfg = c
	return &opts
}

func (o *handlerOptions) TagOptions() models.TagOptions {
	return o.tagOptions
}

func (o *handlerOptions) SetTagOptions(tags models.TagOptions) HandlerOptions {
	opts := *o
	opts.tagOptions = tags
	return &opts
}

func (o *handlerOptions) TimeoutOpts() *prometheus.TimeoutOpts {
	return o.timeoutOpts
}

func (o *handlerOptions) SetTimeoutOpts(t *prometheus.TimeoutOpts) HandlerOptions {
	opts := *o
	opts.timeoutOpts = t
	return &opts
}

func (o *handlerOptions) Enforcer() cost.ChainedEnforcer {
	return o.enforcer
}

func (o *handlerOptions) SetEnforcer(e cost.ChainedEnforcer) HandlerOptions {
	opts := *o
	opts.enforcer = e
	return &opts
}

func (o *handlerOptions) FetchOptionsBuilder() handleroptions.FetchOptionsBuilder {
	return o.fetchOptionsBuilder
}

func (o *handlerOptions) SetFetchOptionsBuilder(
	b handleroptions.FetchOptionsBuilder) HandlerOptions {
	opts := *o
	opts.fetchOptionsBuilder = b
	return &opts
}

func (o *handlerOptions) QueryContextOptions() models.QueryContextOptions {
	return o.queryContextOptions
}

func (o *handlerOptions) SetQueryContextOptions(
	q models.QueryContextOptions) HandlerOptions {
	opts := *o
	opts.queryContextOptions = q
	return &opts
}

func (o *handlerOptions) CPUProfileDuration() time.Duration {
	return o.cpuProfileDuration
}

func (o *handlerOptions) SetCPUProfileDuration(
	c time.Duration) HandlerOptions {
	opts := *o
	opts.cpuProfileDuration = c
	return &opts
}

func (o *handlerOptions) PlacementServiceNames() []string {
	return o.placementServiceNames
}

func (o *handlerOptions) SetPlacementServiceNames(
	n []string) HandlerOptions {
	opts := *o
	opts.placementServiceNames = n
	return &opts
}

func (o *handlerOptions) ServiceOptionDefaults() []handleroptions.ServiceOptionsDefault {
	return o.serviceOptionDefaults
}

func (o *handlerOptions) SetServiceOptionDefaults(
	s []handleroptions.ServiceOptionsDefault) HandlerOptions {
	opts := *o
	opts.serviceOptionDefaults = s
	return &opts
}

func (o *handlerOptions) InstrumentOpts() instrument.Options {
	return o.instrumentOpts
}

func (o *handlerOptions) SetInstrumentOpts(opts instrument.Options) HandlerOptions {
	options := *o
	options.instrumentOpts = opts
	return &options
}

func (o *handlerOptions) NowFn() clock.NowFn {
	return o.nowFn
}

func (o *handlerOptions) SetNowFn(n clock.NowFn) HandlerOptions {
	options := *o
	options.nowFn = n
	return &options
}
